""" Helpers for developping quick test scripts

Creation motivated by fear of filling disc space during long-running stress
tests

.. codeauthor:: Christian Herdtweck, christian.herdtweck@intra2net.com
"""

import contextlib
from threading import Thread
from time import sleep
from file_helpers import get_filesystem_fill_states
from datetime import datetime as dt
from buffers import LogarithmicBuffer
from itertools import tee

class DiscFillCheckerThread(Thread):
    """ a thread that checks disc fill in regular intervals """

    def __init__(self, interval=10, decision_function = None):
        super(DiscFillCheckerThread, self).__init__(name='discFillChck',
                                                    daemon=True)
        # set variables
        self.interval = interval
        if decision_function is None:
            self.decision_function = default_disc_full_decision_function
        else:
            # need to check warn_func somehow (using introspection?)
            self.decision_function = decision_function

        # remember relevant disc stats at start
        self._bufs = {}
        for fs_state in get_filesystem_fill_states():
            self._internal_state_buffer(fs_state)

    def _internal_state_buffer(self, fs_state):
        """ update internal stats buffer, returns all estims

        internal helper called from __init__ and run
        """
        buf = None
        try:
            buf = self._bufs[fs_state.name]
        except KeyError:
            # new file system -- need to create a new buffer
            buf = LogarithmicBuffer(5)
            self._bufs[fs_state.name] =  buf

        buf.add((dt.now(), float(fs_state.available)))
        return buf.get_all()

    def run(self):
        while True:
            sleep(self.interval)
            fs_states = get_filesystem_fill_states()
            for fs_state in fs_states:
                fill_states = self._internal_state_buffer(fs_state)

                # estimate time until full (i.e. when available == 0)
                times_until_empty = \
                    [calc_times_until_zero(old_state, new_state)
                     for old_state, new_state in pairwise(fill_states)]

                # call user-defined function to decide
                min_time = None
                if any(times_until_empty > 0):
                    min_time = min(time for time in times_until_empty
                                   if time>0)
                avg_time = calc_times_until_zero(fill_states[0],
                                                 fill_states[-1])
                self.decision_function(stat,
                                       times_until_empty[-1],     # newest
                                       min(times_until_empty),    # smallest
                                       avg_time)                  # average


def default_disc_full_decision_function(curr_state,
                                        estim_empty_newest,
                                        estim_empty_smallest,
                                        estim_empty_mean):
    """ decide whether to warn or raise error because disc is getting full

    called by DiscFillCheckerThread regularly; params estim_empty_* are result
    of :py:func:`calc_times_until_zero`, so None means that disc state did not
    change.

    :param estim_empty_newest: the newest time estimate
    :param estim_empty_smallest: the smallest of all non-negative estimates
    :param estim_empty_mean: estim between earliest and newest estimate
    """
    pass

def calc_times_until_zero(old_state, new_state):
    """ calc time until value hits zero given (time1, value1), (time2, value2)

    returns time from now until empty in seconds

    delta_date = new_date - old_date
    delta_free = new_free - old_free
    free(t) = new_free + (t-new_t) * delta_free / delta_date
            = 0   <==>
    diff_t := new_t - t = new_free * delta_date / delta_free

    compare to now:
    diff_to_now = t + (-new_t + new_t) - now = (t - new_t) + (new_t - now)
                = -diff_t - (now-new_t)

    To avoid zero-division, returns None if delta_free == 0
    """
    old_date, old_free = old_state
    new_date, new_free = new_state
    if new_free == old_free:
        return None            # avoid zero-division
    time_diff = new_free * (new_date - old_date).total_seconds() \
                         / (old_free - new_free)

    # compare to now
    return (now - new_date).total_seconds() - time_diff


def pairwise(iterable):
    """ s -> (s0,s1), (s1,s2), (s2, s3), ...

    taken from itertool recipes in python api doc of itertools
    """
    a, b = tee(iterable)
    next(b, None)
    return zip(a, b)


@contextlib.contextmanager
def disc_fill_checked(path):
    """ run test function while separate thread watches disc space """

    # todo: add args and forward
    DiscFillCheckerThread().start()

    try:
        yield
    finally:
        pass # wait for other to stop or so? or stop it?
